import typing as t

import numpy as np
import numpy.typing as npt

def rgb_to_hls_(r: int, g: int, b: int) -> tuple[float, float, float]: ...
def rgb_to_hsv_(r: int, g: int, b: int) -> tuple[float, float, float]: ...
def rgb_to_lab_(r: int, g: int, b: int) -> tuple[float, float, float]: ...
def rgb_to_xyz_(r: int, g: int, b: int) -> tuple[float, float, float]: ...
def hls_to_rgb_(h: float, l: float, s: float) -> tuple[int, int, int]: ...
def hsv_to_rgb_(h: float, s: float, v: float) -> tuple[int, int, int]: ...
def lab_to_rgb_(l: float, a: float, b: float) -> tuple[int, int, int]: ...
def xyz_to_rgb_(x: float, y: float, z: float) -> tuple[int, int, int]: ...
def lab_to_xyz_(l: float, a: float, b: float) -> tuple[float, float, float]: ...
def xyz_to_lab_(x: float, y: float, z: float) -> tuple[float, float, float]: ...
def ind_to_rgb_(index: int) -> tuple[int, int, int]: ...
def rgb_to_ind_(r: int, g: int, b: int) -> int: ...
@t.overload
def rgb_to_hls(r: int, g: int = ..., b: int = ...) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_hls(r: tuple[int, int, int]) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_hls(
    r: tuple[int, int, int], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_hls(r: npt.NDArray[np.uint8]) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_hls(
    r: npt.NDArray[np.uint8], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_hls(
    r: npt.NDArray[np.uint8] | tuple[int, int, int] | int,
    g: None | int = ...,
    b: None | int = ...,
) -> tuple[float, float, float] | npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_hsv(r: int, g: int = ..., b: int = ...) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_hsv(r: tuple[int, int, int]) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_hsv(
    r: tuple[int, int, int], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_hsv(r: npt.NDArray[np.uint8]) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_hsv(
    r: npt.NDArray[np.uint8], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_hsv(
    r: npt.NDArray[np.uint8] | tuple[int, int, int] | int, g: None | int = ..., b: None | int = ...
) -> tuple[float, float, float] | npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_xyz(r: int, g: int = ..., b: int = ...) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_xyz(r: tuple[int, int, int]) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_xyz(
    r: tuple[int, int, int], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_xyz(r: npt.NDArray[np.uint8]) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_xyz(
    r: npt.NDArray[np.uint8], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_xyz(
    r: npt.NDArray[np.uint8] | tuple[int, int, int] | int, g: None | int = ..., b: None | int = ...
) -> tuple[float, float, float] | npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_lab(r: int, g: int = ..., b: int = ...) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_lab(r: tuple[int, int, int]) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_lab(
    r: tuple[int, int, int], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> tuple[float, float, float]: ...
@t.overload
def rgb_to_lab(r: npt.NDArray[np.uint8]) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_lab(
    r: npt.NDArray[np.uint8], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> npt.NDArray[np.float_]: ...
@t.overload
def rgb_to_lab(
    r: npt.NDArray[np.uint8] | tuple[int, int, int] | int,
    g: None | int = ...,
    b: None | int = ...,
) -> tuple[float, float, float] | npt.NDArray[np.float_]: ...
@t.overload
def xyz_to_lab(x: float, y: float = ..., z: float = ...) -> tuple[float, float, float]: ...
@t.overload
def xyz_to_lab(x: tuple[float, float, float]) -> tuple[float, float, float]: ...
@t.overload
def xyz_to_lab(
    x: tuple[float, float, float], y: t.Literal[None] = ..., z: t.Literal[None] = ...
) -> tuple[float, float, float]: ...
@t.overload
def xyz_to_lab(x: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
@t.overload
def xyz_to_lab(
    x: npt.NDArray[np.float_], y: t.Literal[None] = ..., z: t.Literal[None] = ...
) -> npt.NDArray[np.float_]: ...
@t.overload
def lab_to_xyz(l: float, a: float = ..., b: float = ...) -> tuple[float, float, float]: ...
@t.overload
def lab_to_xyz(l: tuple[float, float, float]) -> tuple[float, float, float]: ...
@t.overload
def lab_to_xyz(
    l: tuple[float, float, float], a: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> tuple[float, float, float]: ...
@t.overload
def lab_to_xyz(l: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]: ...
@t.overload
def lab_to_xyz(
    l: npt.NDArray[np.float_], a: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> npt.NDArray[np.float_]: ...
@t.overload
def hls_to_rgb(h: float, l: float = ..., s: float = ...) -> tuple[int, int, int]: ...
@t.overload
def hls_to_rgb(h: tuple[float, float, float]) -> tuple[int, int, int]: ...
@t.overload
def hls_to_rgb(
    h: tuple[float, float, float], l: t.Literal[None] = ..., s: t.Literal[None] = ...
) -> tuple[int, int, int]: ...
@t.overload
def hls_to_rgb(h: npt.NDArray[np.float_]) -> npt.NDArray[np.uint8]: ...
@t.overload
def hls_to_rgb(
    h: npt.NDArray[np.float_], l: t.Literal[None] = ..., s: t.Literal[None] = ...
) -> npt.NDArray[np.uint8]: ...
@t.overload
def hsv_to_rgb(h: float, s: float = ..., v: float = ...) -> tuple[int, int, int]: ...
@t.overload
def hsv_to_rgb(h: tuple[float, float, float]) -> tuple[int, int, int]: ...
@t.overload
def hsv_to_rgb(
    h: tuple[float, float, float], s: t.Literal[None] = ..., v: t.Literal[None] = ...
) -> tuple[int, int, int]: ...
@t.overload
def hsv_to_rgb(h: npt.NDArray[np.float_]) -> npt.NDArray[np.uint8]: ...
@t.overload
def hsv_to_rgb(
    h: npt.NDArray[np.float_], s: t.Literal[None] = ..., v: t.Literal[None] = ...
) -> npt.NDArray[np.uint8]: ...
@t.overload
def xyz_to_rgb(x: float, y: float = ..., z: float = ...) -> tuple[int, int, int]: ...
@t.overload
def xyz_to_rgb(x: tuple[float, float, float]) -> tuple[int, int, int]: ...
@t.overload
def xyz_to_rgb(
    x: tuple[float, float, float], y: t.Literal[None] = ..., z: t.Literal[None] = ...
) -> tuple[int, int, int]: ...
@t.overload
def xyz_to_rgb(x: npt.NDArray[np.float_]) -> npt.NDArray[np.uint8]: ...
@t.overload
def xyz_to_rgb(
    x: npt.NDArray[np.float_], y: t.Literal[None] = ..., z: t.Literal[None] = ...
) -> npt.NDArray[np.uint8]: ...
@t.overload
def lab_to_rgb(l: float, a: float = ..., b: float = ...) -> tuple[int, int, int]: ...
@t.overload
def lab_to_rgb(l: tuple[float, float, float]) -> tuple[int, int, int]: ...
@t.overload
def lab_to_rgb(
    l: tuple[float, float, float], a: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> tuple[int, int, int]: ...
@t.overload
def lab_to_rgb(l: npt.NDArray[np.float_]) -> npt.NDArray[np.uint8]: ...
@t.overload
def lab_to_rgb(
    l: npt.NDArray[np.float_], a: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> npt.NDArray[np.uint8]: ...
@t.overload
def ind_to_rgb(index: int) -> tuple[int, int, int]: ...
@t.overload
def ind_to_rgb(index: npt.NDArray[np.uint32]) -> npt.NDArray[np.uint8]: ...
@t.overload
def rgb_to_ind(r: int, g: int = ..., b: int = ...) -> int: ...
@t.overload
def rgb_to_ind(r: tuple[int, int, int]) -> int: ...
@t.overload
def rgb_to_ind(
    r: tuple[int, int, int], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> int: ...
@t.overload
def rgb_to_ind(r: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint32]: ...
@t.overload
def rgb_to_ind(
    r: npt.NDArray[np.uint8], g: t.Literal[None] = ..., b: t.Literal[None] = ...
) -> npt.NDArray[np.uint32]: ...
@t.overload
def rgb_to_ind(
    r: npt.NDArray[np.uint8] | tuple[int, int, int] | int,
    g: tuple[int, int, int] | None | int = ...,
    b: tuple[int, int, int] | None | int = ...,
) -> int | npt.NDArray[np.uint32]: ...



import typing as t

from numba import prange as prange
from numba.core import types as types

_P = t.ParamSpec("_P")
_R = t.TypeVar("_R")

@t.overload
def njit(*args: t.Callable[_P, _R]) -> t.Callable[_P, _R]: ...  # type: ignore[misc]
@t.overload
def njit(**kwargs: t.Any) -> t.Callable[[t.Callable[_P, _R]], t.Callable[_P, _R]]: ...
@t.overload
def njit(
    *args: list[t.Any], **kwargs: t.Any
) -> t.Callable[[t.Callable[_P, _R]], t.Callable[_P, _R]]: ...
@t.overload
def njit(
    *args: t.Callable[_P, _R] | list[t.Any], **kwargs: t.Any
) -> t.Callable[[t.Callable[_P, _R]], t.Callable[_P, _R]] | t.Callable[_P, _R]: ...
@t.overload
def register_jitable(*args: t.Callable[_P, _R]) -> t.Callable[_P, _R]: ...  # type: ignore[misc]
@t.overload
def register_jitable(**kwargs: t.Any) -> t.Callable[[t.Callable[_P, _R]], t.Callable[_P, _R]]: ...
@t.overload
def register_jitable(
    *args: list[t.Any], **kwargs: t.Any
) -> t.Callable[[t.Callable[_P, _R]], t.Callable[_P, _R]]: ...
@t.overload
def register_jitable(
    *args: t.Callable[_P, _R] | list[t.Any], **kwargs: t.Any
) -> t.Callable[[t.Callable[_P, _R]], t.Callable[_P, _R]] | t.Callable[_P, _R]: ...



# type: ignore
# ruff: noqa: E741
"""
This contains low level conversion functions.

Adapted from https://github.com/CairX/convert-colors-py
https://www.easyrgb.com/en/math.php
"""

# fmt: off

import numpy as np
from numba.core.errors import NumbaTypeError
from numba.extending import overload

from jccc._numba import njit, prange, register_jitable, types

_1_DIV_255: float =                           1.0 / 255.0
_1_DIV_360: float =                           1.0 / 360.0
_1_DIV_100: float =                           1.0 / 100.0
_1_DIV_1POINT055: float =                     1.0 / 1.055
_1_DIV_12POINT92: float =                     1.0 / 12.92
_1_DIV_95_047: float =                        1.0 / 95.047
_1_DIV_108_883: float =                       1.0 / 108.883
_1_DIV_116: float =                           1.0 / 116.0
_1_DIV_500: float =                           1.0 / 500.0
_1_DIV_200: float =                           1.0 / 200.0
_1_DIV_903_3: float =                         1.0 / 903.3
_1_THIRD: float =                             1.0 / 3.0
_1_SIXTH: float =                             1.0 / 6.0
_2_THIRDS: float =                            2.0 / 3.0

# fmt: on


def conversion_overload(overload_func, **overload_kwargs):
    """
    Decorator that provides the overloads for RGB conversions that result in
    float triplets (e.g., xyz, hls, hsv, etc.). The signature of the
    `overload_func` MUST be the following:

    def func_name(e, f, g):
        ...

    The base_implementation_func must have a three argument signature.
    """
    # no nans, no infs, and no signed zeros
    overload_kwargs["fastmath"] = overload_kwargs.pop("fastmath", {"nnan", "nsz", "ninf"})

    func_name = overload_func.__name__
    if func_name in ("_rgb_to_ind", "_ind_to_rgb"):
        # Most of the functions maintain the same array structure and pivot from
        # uint8 to float32 or visa versa. However, rgb_to_ind and ind_to_rgb,
        # remove and add dimensions to the outpput and pivot between uint8 and
        # uint32. So, those functions are handled separately
        return index_conversions_overloads(overload_func, **overload_kwargs)

    if "to_rgb" in func_name:
        itype = types.Float
        dtype = np.uint8

    elif "rgb_to" in func_name:
        itype = types.Integer
        dtype = np.float32

    else:
        itype = types.Float
        dtype = np.float32

    def inner(base_implementation_func):
        # make sure the base implementation can be run in a jitted context
        base_implementation_func = register_jitable(**overload_kwargs)(base_implementation_func)

        @overload(overload_func, **overload_kwargs)
        def base_impl(e, f, g):
            # implementation when all three arguments are itype
            if isinstance(e, itype) and isinstance(f, itype) and isinstance(g, itype):
                # ensure signature matches without forcing the
                # base_implementation_func's signature
                return lambda e, f, g: base_implementation_func(e, f, g)

            raise NumbaTypeError("NumbaTypeError")

        @overload(overload_func, **overload_kwargs)
        def tuple_impl(e, f, g):
            # implementation when first argument is a tuple
            if isinstance(e, types.UniTuple) and len(e) == 3:
                return lambda e, f, g: base_implementation_func(e[0], e[1], e[2])
            raise NumbaTypeError("NumbaTypeError")

        @overload(overload_func, **overload_kwargs)
        def array_impl(e, f, g):
            # implementation for arrays
            if isinstance(e, types.Array) and isinstance(e.dtype, itype):
                if e.ndim == 1:

                    def impl1D(e, f, g):
                        return np.array(
                            base_implementation_func(e[0], e[1], e[2]),
                            dtype=dtype,
                        )

                    return impl1D

                elif e.ndim == 2:

                    def impl2D(e, f, g):
                        height, _ = e.shape
                        out = np.zeros_like(e, dtype=dtype)
                        for i in prange(height):
                            out[i] = base_implementation_func(
                                e[i, 0],
                                e[i, 1],
                                e[i, 2],
                            )
                        return out

                    return impl2D
                elif e.ndim == 3:

                    def impl3D(e, f, g):
                        height, width, _ = e.shape
                        out = np.zeros_like(e, dtype=dtype)
                        for i in prange(height):
                            for j in range(width):
                                out[i, j] = base_implementation_func(
                                    e[i, j, 0],
                                    e[i, j, 1],
                                    e[i, j, 2],
                                )
                        return out

                    return impl3D
                elif e.ndim == 4:

                    def impl4D(e, f, g):
                        n_frames, height, width, _ = e.shape
                        out = np.zeros_like(e, dtype=dtype)
                        for i in prange(n_frames):
                            for j in range(height):
                                for k in range(width):
                                    out[i, j, k] = base_implementation_func(
                                        e[i, j, k, 0],
                                        e[i, j, k, 1],
                                        e[i, j, k, 2],
                                    )
                        return out

                    return impl4D

            raise NumbaTypeError("NumbaTypeError")

        return base_implementation_func

    return inner


def index_conversions_overloads(overload_func, **overload_kwargs):
    """
    The signature of the `overload_func` MUST be the following:

    def func_name(e, f, g):
        ...

    """
    func_name = overload_func.__name__
    if func_name == "_ind_to_rgb":
        itype = types.Integer
        fgtype = types.NoneType

    elif func_name == "_rgb_to_ind":
        itype = fgtype = types.Integer

    else:
        raise ValueError("Invalid function")

    def inner(base_implementation_func):
        # make sure the base implementation can be run in a jitted context
        base_implementation_func = register_jitable(**overload_kwargs)(base_implementation_func)

        @overload(overload_func, **overload_kwargs)
        def base_impl(e, f, g):
            # implementation when all three arguments are itype
            if isinstance(e, itype) and isinstance(f, fgtype) and isinstance(g, fgtype):
                # ensure signature matches without forcing the
                # base_implementation_func's signature
                return lambda e, f, g: base_implementation_func(e, f, g)
            raise NumbaTypeError("NumbaTypeError")

        # NO TUPLE IMPLEMENTATION
        if func_name == "_rgb_to_ind":

            @overload(overload_func, **overload_kwargs)
            def array_impl(e, f, g):
                # implementation for arrays
                if isinstance(e, types.Array) and isinstance(e.dtype, itype):
                    if 1 < e.ndim <= 4:

                        def impl(e, f, g):
                            return (
                                np.left_shift(e[..., 0], 16)
                                + np.left_shift(e[..., 1], 8)
                                + e[..., 2]
                            ).astype(np.uint32)

                        return impl

                    elif e.ndim == 1:

                        def impl(e, f, g):
                            return np.array(
                                [base_implementation_func(e[0], e[1], e[2])],
                                dtype=np.uint32,
                            )

                        return impl

                raise NumbaTypeError("NumbaTypeError")

        else:

            @overload(overload_func, **overload_kwargs)
            def array_impl(e, f, g):
                # implementation for arrays
                if isinstance(e, types.Array) and isinstance(e.dtype, itype) and (1 <= e.ndim <= 3):

                    def impl(e, f, g):
                        out = np.zeros(e.shape + (3,), dtype=np.uint8)
                        out[..., 0] = np.bitwise_and(np.right_shift(e, 16), 255)
                        out[..., 1] = np.bitwise_and(np.right_shift(e, 8), 255)
                        out[..., 2] = np.bitwise_and(e, 255)
                        return out

                    return impl

                raise NumbaTypeError("NumbaTypeError")

        return base_implementation_func

    return inner


################################################################################
### RGB --> HLS
################################################################################


def _rgb_to_hls(e, f, g):
    raise NotImplementedError


@conversion_overload(_rgb_to_hls, inline="always", cache=True)
def rgb_to_hls_(r, g, b):
    r *= _1_DIV_255
    g *= _1_DIV_255
    b *= _1_DIV_255
    minc, maxc = min(r, g, b), max(r, g, b)
    sumc = maxc + minc
    l = sumc / 2.0
    if minc == maxc:
        return 0.0, l * 100.0, 0.0

    rangec = maxc - minc
    if l <= 0.5:
        s = rangec / sumc
    else:
        s = rangec / (2.0 - sumc)
    rc = (maxc - r) / rangec
    gc = (maxc - g) / rangec
    bc = (maxc - b) / rangec
    if r == maxc:
        h = bc - gc
    elif g == maxc:
        h = 2.0 + rc - bc
    else:
        h = 4.0 + gc - rc
    h = (h / 6.0) % 1.0

    h *= 360.0
    l *= 100.0
    s *= 100.0
    return h, l, s


@njit(inline="always", cache=True)
def rgb_to_hls(r, g=None, b=None):
    return _rgb_to_hls(r, g, b)


################################################################################
### HLS --> RGB
################################################################################


def _hls_to_rgb(e, f, g):
    raise NotImplementedError


@conversion_overload(_hls_to_rgb, inline="always", cache=True)
def hls_to_rgb_(h, l, s):
    h *= _1_DIV_360
    l *= _1_DIV_100
    s *= _1_DIV_100
    if s == 0.0:
        l *= 255.0
        r = g = b = int(round(l))
        return r, g, b
    if l <= 0.5:
        m2 = l * (1.0 + s)
    else:
        m2 = l + s - (l * s)
    m1 = 2.0 * l - m2
    r, g, b = (_v(m1, m2, h + _1_THIRD), _v(m1, m2, h), _v(m1, m2, h - _1_THIRD))

    r *= 255.0
    g *= 255.0
    b *= 255.0
    return int(round(r)), int(round(g)), int(round(b))


@register_jitable(inline="always", cache=True)
def _v(m1, m2, hue):
    hue = hue % 1.0
    if hue < _1_SIXTH:
        return m1 + (m2 - m1) * hue * 6.0
    if hue < 0.5:
        return m2
    if hue < _2_THIRDS:
        return m1 + (m2 - m1) * (_2_THIRDS - hue) * 6.0
    return m1


@njit(inline="always", cache=True)
def hls_to_rgb(h, l=None, s=None):
    return _hls_to_rgb(h, l, s)


################################################################################
### RGB --> HSV
################################################################################


def _rgb_to_hsv(e, f, g):
    raise NotImplementedError


@conversion_overload(_rgb_to_hsv, inline="always", cache=True)
def rgb_to_hsv_(r, g, b):
    r *= _1_DIV_255
    g *= _1_DIV_255
    b *= _1_DIV_255
    minc, maxc = min(r, g, b), max(r, g, b)
    v = maxc
    if minc == maxc:
        return 0.0, 0.0, v * 100.0

    rangec = maxc - minc
    s = rangec / maxc

    # rc = (maxc-r) / rangec
    # gc = (maxc-g) / rangec
    # bc = (maxc-b) / rangec
    denom = 1.0 / rangec
    if r == maxc:
        # h = bc-gc
        h = ((maxc - b) * denom) - ((maxc - g) * denom)
    elif g == maxc:
        # h = 2.0+rc-bc
        h = 2.0 + ((maxc - r) * denom) - ((maxc - b) * denom)
    else:
        # h = 4.0+gc-rc
        h = 4.0 + ((maxc - g) * denom) - ((maxc - r) * denom)

    h = (h / 6.0) % 1.0

    h *= 360.0
    s *= 100.0
    v *= 100.0
    return h, s, v


@njit(inline="always", cache=True)
def rgb_to_hsv(r, g=None, b=None):
    return _rgb_to_hsv(r, g, b)


################################################################################
### HSV --> RGB
################################################################################


def _hsv_to_rgb(e, f, g):
    raise NotImplementedError


@conversion_overload(_hsv_to_rgb, inline="always", cache=True)
def hsv_to_rgb_(h, s, v):
    h *= _1_DIV_360
    s *= _1_DIV_100
    v *= _1_DIV_100
    if s == 0.0:
        r = g = b = int(round(v * 100.0))
        return r, g, b

    i = int(h * 6.0)  # XXX assume int() truncates!
    f = (h * 6.0) - i
    p = v * (1.0 - s)
    q = v * (1.0 - s * f)
    t = v * (1.0 - s * (1.0 - f))
    i = i % 6

    if i == 0:
        r, g, b = v, t, p

    elif i == 1:
        r, g, b = q, v, p

    elif i == 2:
        r, g, b = p, v, t

    elif i == 3:
        r, g, b = p, q, v

    elif i == 4:
        r, g, b = t, p, v

    else:
        r, g, b = v, p, q

    r *= 255.0
    g *= 255.0
    b *= 255.0
    return int(round(r)), int(round(g)), int(round(b))


@njit(inline="always", cache=True)
def hsv_to_rgb(h, s=None, v=None):
    return _hsv_to_rgb(h, s, v)


################################################################################
### RGB --> XYZ
################################################################################


def _rgb_to_xyz(e, f, g):
    raise NotImplementedError


@conversion_overload(_rgb_to_xyz, inline="always", cache=True)
def rgb_to_xyz_(r, g, b):
    r = _pivot_rgb_to_xyz(r * _1_DIV_255)
    g = _pivot_rgb_to_xyz(g * _1_DIV_255)
    b = _pivot_rgb_to_xyz(b * _1_DIV_255)

    x = r * 0.4124564 + g * 0.3575761 + b * 0.1804375
    y = r * 0.2126729 + g * 0.7151522 + b * 0.0721750
    z = r * 0.0193339 + g * 0.1191920 + b * 0.9503041

    return x, y, z


@register_jitable([types.float64(types.float64)], cache=True)
def _pivot_rgb_to_xyz(value):
    if value <= 0.04045:
        value *= _1_DIV_12POINT92
    else:
        value = ((value + 0.055) * _1_DIV_1POINT055) ** 2.4
    return value * 100.0


@njit(inline="always", cache=True)
def rgb_to_xyz(r, g=None, b=None):
    """
    Convert tuple from the sRGB color space to the CIE XYZ color space.

    The XYZ output is determined using D65 illuminate with a 2° observer angle.
    https://en.wikipedia.org/wiki/Illuminant_D65

    The conversion matrix used was provided by Bruce Lindbloom:
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html

    Formulas for conversion:
    http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html
    https://easyrgb.com/en/math.php

    Information about respective color space:
    sRGB (standard Red Green Blue): https://en.wikipedia.org/wiki/SRGB
    CIE XYZ: https://en.wikipedia.org/wiki/CIE_1931_color_space
    """
    return _rgb_to_xyz(r, g, b)


################################################################################
### XYZ --> RGB
################################################################################


def _xyz_to_rgb(e, f, g):
    raise NotImplementedError


@conversion_overload(_xyz_to_rgb, inline="always", cache=True)
def xyz_to_rgb_(x, y, z):
    x *= _1_DIV_100
    y *= _1_DIV_100
    z *= _1_DIV_100

    r = x * 3.2404542 + y * -1.5371385 + z * -0.4985314
    g = x * -0.9692660 + y * 1.8760108 + z * 0.0415560
    b = x * 0.0556434 + y * -0.2040259 + z * 1.0572252

    r = _pivot_xyz_to_rgb(r)
    g = _pivot_xyz_to_rgb(g)
    b = _pivot_xyz_to_rgb(b)

    return r, g, b


@register_jitable([types.uint8(types.float64)], cache=True)
def _pivot_xyz_to_rgb(value):
    if value <= 0.0031308:
        value *= 12.92
    else:
        value = ((value**0.4166666) * 1.055) - 0.055
    return int(round(value * 255.0))


@njit(inline="always", cache=True)
def xyz_to_rgb(x, y=None, z=None):
    """
    Convert tuple from the CIE XYZ color space to the sRGB color space.

    Conversion is based on that the XYZ input uses an the D65 illuminate with a 2° observer angle.
    https://en.wikipedia.org/wiki/Illuminant_D65

    The inverse conversion matrix used was provided by Bruce Lindbloom:
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html

    Formulas for conversion:
    http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html
    https://easyrgb.com/en/math.php

    Information about respective color space:
    sRGB (standard Red Green Blue): https://en.wikipedia.org/wiki/SRGB
    CIE XYZ: https://en.wikipedia.org/wiki/CIE_1931_color_space
    """
    return _xyz_to_rgb(x, y, z)


################################################################################
### XYZ --> LAB
################################################################################


def _xyz_to_lab(e, f, g):
    raise NotImplementedError


@conversion_overload(_xyz_to_lab, inline="always", cache=True)
def xyz_to_lab_(x, y, z):
    x = _pivot_xyz_to_lab(x * _1_DIV_95_047)
    y = _pivot_xyz_to_lab(y * _1_DIV_100)
    z = _pivot_xyz_to_lab(z * _1_DIV_108_883)

    l = max(0.0, (116.0 * y) - 16.0)
    a = (x - y) * 500.0
    b = (y - z) * 200.0

    return l, a, b


@register_jitable([types.float64(types.float64)], cache=True)
def _pivot_xyz_to_lab(value):
    if value > 0.008856:
        value = value**_1_THIRD
    else:
        value = ((value * 903.3) + 16.0) * _1_DIV_116
    return value


@njit(inline="always", cache=True)
def xyz_to_lab(x, y=None, z=None):
    """
    Convert tuple from the CIE XYZ color space to the CIE L*a*b color space.

    Conversion is based on that the XYZ input uses an the D65 illuminate with a 2° observer angle.
    https://en.wikipedia.org/wiki/Illuminant_D65

    Formulas for conversion:
    https://en.wikipedia.org/wiki/CIELAB_color_space#CIELAB%E2%80%93CIEXYZ_conversions
    http://www.brucelindbloom.com/index.html?Eqn_XYZ_to_Lab.html
    https://easyrgb.com/en/math.php

    Information about respective color space:
    CIE L*a*b: https://en.wikipedia.org/wiki/Lab_color_space
    CIE XYZ: https://en.wikipedia.org/wiki/CIE_1931_color_space
    """
    return _xyz_to_lab(x, y, z)


################################################################################
### LAB --> XYZ
################################################################################


def _lab_to_xyz(e, f, g):
    raise NotImplementedError


@conversion_overload(_lab_to_xyz, inline="always", cache=True)
def lab_to_xyz_(l, a, b):
    # Reminder: The y values is calculated first as it can be reused
    # for the calculation of x and z.
    y = (l + 16.0) * _1_DIV_116
    x = y + (a * _1_DIV_500)
    z = y - (b * _1_DIV_200)

    x3 = x**3
    z3 = z**3

    x = x3 if x3 > 0.008856 else ((x * 116.0) - 16.0) * _1_DIV_903_3
    y = (y**3) if l > 7.9996248 else l * _1_DIV_903_3
    z = z3 if z3 > 0.008856 else ((z * 116.0) - 16.0) * _1_DIV_903_3

    x *= 95.047
    y *= 100.000
    z *= 108.883

    return x, y, z


@njit(inline="always", cache=True)
def lab_to_xyz(l, a=None, b=None):
    """
    Convert tuple from the CIE L*a*b* color space to the CIE XYZ color space.

    The XYZ output is determined using D65 illuminate with a 2° observer angle.
    https://en.wikipedia.org/wiki/Illuminant_D65

    Formulas for conversion:
    https://en.wikipedia.org/wiki/CIELAB_color_space#CIELAB%E2%80%93CIEXYZ_conversions
    http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html
    https://easyrgb.com/en/math.php

    Information about respective color space:
    CIE L*a*b: https://en.wikipedia.org/wiki/Lab_color_space
    CIE XYZ: https://en.wikipedia.org/wiki/CIE_1931_color_space
    """
    return _lab_to_xyz(l, a, b)


################################################################################
### RGB --> LAB
################################################################################


def _rgb_to_lab(e, f, g):
    raise NotImplementedError


@conversion_overload(_rgb_to_lab, inline="always", cache=True)
def rgb_to_lab_(r, g, b):
    xyz = rgb_to_xyz_(r, g, b)
    return xyz_to_lab_(xyz[0], xyz[1], xyz[2])


@njit(inline="always", cache=True)
def rgb_to_lab(r, g=None, b=None):
    """
    Convert tuple from the sRGB color space to the CIE L*a*b* color space.
    Shorthand method for chaining sRGB => CIE XYZ => CIE L*a*b*.
    """
    return _rgb_to_lab(r, g, b)


################################################################################
### LAB --> RGB
################################################################################


def _lab_to_rgb(e, f, g):
    raise NotImplementedError


@conversion_overload(_lab_to_rgb, inline="always", cache=True)
def lab_to_rgb_(l, a, b):
    xyz = lab_to_xyz_(l, a, b)
    return xyz_to_rgb_(xyz[0], xyz[1], xyz[2])


@njit(inline="always", cache=True)
def lab_to_rgb(l, a=None, b=None):
    """
    Convert tuple from the CIE L*a*b* color space to the sRGB color space.
    Shorthand method for chaining CIE L*a*b* => CIE XYZ  => sRGB.
    """
    return _lab_to_rgb(l, a, b)


################################################################################
### RGB --> Indices
################################################################################


def _rgb_to_ind(e, f, g):
    raise NotImplementedError


@conversion_overload(_rgb_to_ind, inline="always", cache=True)
def rgb_to_ind_(r, g, b):
    return (r << 16) + (g << 8) + b


@njit(inline="always", cache=True)
def rgb_to_ind(r, g=None, b=None):
    """
    Uses bit shifting to get the index for an RGB triplet.
    """
    return _rgb_to_ind(r, g, b)


################################################################################
### Indices --> RGB
################################################################################


def _ind_to_rgb(e, f, g):
    raise NotImplementedError


@conversion_overload(_ind_to_rgb, inline="always", cache=True)
def ind_to_rgb_(e, f, g):
    return (e >> 16) & 255, (e >> 8) & 255, e & 255


@njit(inline="always", cache=True)
def ind_to_rgb(index):
    """
    Uses bit shifting to get the RGB triplet from an index.
    """
    return _ind_to_rgb(index, None, None)



import os as _os

from numba import njit as njit  # type: ignore
from numba import prange as prange
from numba.core import types as types  # type: ignore
from numba.extending import register_jitable as register_jitable  # type: ignore

# https://numba.readthedocs.io/en/latest/reference/deprecation.html#deprecation-of-old-style-numba-captured-errors
_os.environ["NUMBA_CAPTURED_ERRORS"] = "new_style"



# fmt: off
import math

from jccc._numba import njit

_25_POW_7: float =         float(6_103_515_625)
_2PI: float =              float(2.0 * math.pi)
_1_DIV_25: float =         float(1.0 / 25.0)
_30_RAD: float =           math.radians(30)
_6_RAD: float =            math.radians(6)
_63_RAD: float =           math.radians(63)
_KL: float =               float(1.0)
_KC: float =               float(1.0)
_KH: float =               float(1.0)

# fmt: on


@njit(inline="always")
def delta_e_cie2000(
    L1: float,
    a1: float,
    b1: float,
    L2: float,
    a2: float,
    b2: float,
    kL: float = _KL,
    kC: float = _KC,
    kH: float = _KH,
) -> float:
    """
    Calculate ΔE between two CIE LAB values.

    see https://www.hajim.rochester.edu/ece/sites/gsharma/ciede2000/ciede2000noteCRNA.pdf
    """
    # distort `a` based on average chroma
    # then convert to lch coordinates from distorted `a`
    # all subsequence calculations are in the new coordinates
    # (often denoted "prime" in the literature)
    Cbar = 0.5 * (math.hypot(a1, b1) + math.hypot(a2, b2))
    c7 = Cbar**7
    G = 0.5 * (1 - math.sqrt(c7 / (c7 + _25_POW_7)))
    scale = 1 + G

    a1scale = a1 * scale
    C1 = math.hypot(a1scale, b1)
    h1: float = math.atan2(b1, a1scale)
    if h1 < 0.0:
        h1 += _2PI

    a2scale = a2 * scale
    C2: float = math.hypot(a2scale, b2)
    h2: float = math.atan2(b2, a2scale)
    if h2 < 0.0:
        h2 += _2PI

    # recall that c, h are polar coordinates.  c==r, h==theta

    # cide2000 has four terms to delta_e:
    # 1) Luminance term
    # 2) Hue term
    # 3) Chroma term
    # 4) hue Rotation term

    # lightness term
    Lbar = 0.5 * (L1 + L2)
    tmp = (Lbar - 50) ** 2
    SL = 1 + 0.015 * tmp / math.sqrt(20 + tmp)
    L_term = (L2 - L1) / (kL * SL)

    # chroma term
    Cbar = 0.5 * (C1 + C2)  # new coordinates
    SC = 1 + 0.045 * Cbar
    C_term = (C2 - C1) / (kC * SC)

    # hue term
    h_diff = h2 - h1
    h_sum = h1 + h2
    CC = C1 * C2

    dH = h_diff
    if CC == 0.0:
        dH = 0.0  # if r == 0, dtheta == 0
    else:
        if h_diff > math.pi:
            dH = h_diff - _2PI
        elif h_diff < -math.pi:
            dH = h_diff + _2PI
        else:
            dH = h_diff

    dH_term = 2.0 * math.sqrt(CC) * math.sin(dH / 2.0)

    if CC == 0.0:
        Hbar = h_sum
    else:
        if math.fabs(h1 - h2) <= math.pi:
            Hbar = h_sum * 0.5
        else:
            if h_sum >= _2PI:
                Hbar = (h_sum - _2PI) * 0.5
            else:
                Hbar = (h_sum + _2PI) * 0.5

    T = (
        1
        - 0.17 * math.cos(Hbar - _30_RAD)
        + 0.24 * math.cos(2 * Hbar)
        + 0.32 * math.cos(3 * Hbar + _6_RAD)
        - 0.20 * math.cos(4 * Hbar - _63_RAD)
    )
    SH = 1 + 0.015 * Cbar * T

    H_term = dH_term / (kH * SH)

    # hue rotation
    c7 = Cbar**7
    Rc = 2 * math.sqrt(c7 / (c7 + _25_POW_7))
    dtheta = _30_RAD * math.exp(-(((math.degrees(Hbar) - 275.0) * _1_DIV_25) ** 2))
    R_term = -math.sin(2.0 * dtheta) * Rc * C_term * H_term

    # put it all together
    dE2 = L_term**2
    dE2 += C_term**2
    dE2 += H_term**2
    dE2 += R_term
    dE2 = math.sqrt(max(dE2, 0.0))
    return dE2



from __future__ import annotations

import typing as t
from pathlib import Path

import numpy as np
import numpy.typing as npt

from jccc.convert import rgb_to_ind, rgb_to_lab
from jccc.utils import create_lookup_table, get_all_lab, get_data_path

Float64Array: t.TypeAlias = npt.NDArray[np.float64]
Uint8Array: t.TypeAlias = npt.NDArray[np.uint8]
Uint16Array: t.TypeAlias = npt.NDArray[np.uint16]
AnyArray: t.TypeAlias = npt.NDArray[t.Any]
StrArray: t.TypeAlias = npt.NDArray[np.str_]

Dtypeish: t.TypeAlias = list[tuple[str, str]]
NumpyLoadable: t.TypeAlias = Path | str | t.Iterable[str]


def _load_data(data: NumpyLoadable | None, dtype: Dtypeish) -> AnyArray:
    """Load the data using numpy."""
    # Validate the dtype
    expects = {"group", "name", "r", "g", "b"}
    actual = {elem[0] for elem in dtype}
    if not expects.issubset(actual):
        err = ValueError()
        err.add_note(f"Missing required columns in `dtype`:\n{sorted(expects - actual):!r}")
        raise err

    # Validate the data
    if data is None:
        raise ValueError("An argument to must be provided for either `file` or `lines")

    data = np.loadtxt(data, dtype=dtype, delimiter=",")
    data = np.sort(data, order=["group", "r", "g", "b"])
    return data


# ============================================================================ #
# BasePalette
# ============================================================================ #


class _BasePalette:
    """
    Base class used by all color palettes.
    """

    def __init__(self, name: t.LiteralString, data: NumpyLoadable | None, dtype: Dtypeish) -> None:
        self._name = name
        self._data = _load_data(data, dtype)
        self._lookup_table: Uint16Array | None = None

    def convert_to_rgbs(self, arr: Uint8Array) -> Uint8Array:
        """
        Converts an array of RGB values in corresponding palette RGBs.
        """
        return self.rgbs[self.convert_to_indices(arr)]

    def convert_to_indices(self, arr: Uint8Array) -> Uint16Array:
        """
        Converts an array of RGB values to the corresponding palette indices.
        This is useful for getting other palette information or just reducing
        memory usage.
        """
        return self.lookup_table[rgb_to_ind(arr)]

    @property
    def lookup_table_path(self) -> Path:
        """
        Return the path to the corresponding lookup table.
        """
        return getattr(self, "_lookup_table_path", Path(f"{self._name}.npz").absolute())

    @property
    def lookup_table(self) -> Uint16Array:
        """
        Returns the lookup table.
        """
        if self._lookup_table is None:
            path = self.lookup_table_path
            if not path.exists():
                err = FileNotFoundError(path)
                err.add_note("Call `write_lookup_table` first.")
                raise err

            self._lookup_table = np.load(path)

        return self._lookup_table

    @lookup_table.setter
    def lookup_table(self, value: Uint16Array | None) -> None:
        self._lookup_table = value

    def write_lookup_table(self, *, overwrite: bool = False, strict: bool = True) -> None:
        """
        Write a lookup table to a path using the working directory and the name
        of the palette.
        """
        path = self.lookup_table_path
        if not overwrite and path.exists():
            if not strict:
                return

            raise FileExistsError(path)

        all_lab = get_all_lab()
        arr_lab = rgb_to_lab(self.rgbs)
        lookup_table = create_lookup_table(arr_lab, all_lab)
        np.savez_compressed(path, lookup_table)

    @property
    def data(self) -> AnyArray:
        """
        Returns the loaded palette data.
        """
        return self._data

    @property
    def rgbs(self) -> Uint8Array:
        """
        Return the `r`, `g`, `b` columns as a stacked array.
        """
        return np.column_stack(tuple(self.data[k] for k in "rgb")).astype(np.uint8)

    @property
    def groups(self) -> StrArray:
        """
        Return the `group` column.
        """
        return self.data["group"]

    @property
    def names(self) -> StrArray:
        """
        Return the `name` column.
        """
        return self.data["name"]

    @property
    def n_colors(self) -> int:
        """
        Returns the number of colors in the palette.
        """
        return self.data.shape[0]


# ============================================================================ #
# Palette


class Palette(_BasePalette):
    """
    This class represents a color palette that can be used for color
    conversions.
    """


# ============================================================================ #
# CSS4 Palette


class Css4Palette(Palette):
    """
    Represents the CSS4 color palette.
    """

    def __init__(self) -> None:
        super().__init__(
            "css4",
            get_data_path("palettes/css4.csv", strict=True),
            [
                ("group", "U20"),
                ("name", "U50"),
                ("r", "u1"),
                ("g", "u1"),
                ("b", "u1"),
            ],
        )
        self._lookup_table_path = get_data_path("palettes/css4.npz")


if __name__ == "__main__":
    palettes = [Css4Palette]
    # construct the palette lookup tables
    for klass in palettes:
        palette = klass()
        print(f"Creating {klass.__name__} lookup table")
        palette.write_lookup_table(strict=False)



from __future__ import annotations

import typing as t
from pathlib import Path

import numpy as np
import numpy.typing as npt

from jccc._numba import njit, prange
from jccc.convert import rgb_to_lab
from jccc.diff import delta_e_cie2000

Float16Array: t.TypeAlias = npt.NDArray[np.float16]
Float32Array: t.TypeAlias = npt.NDArray[np.float32]
Float64Array: t.TypeAlias = npt.NDArray[np.float64]
Uint8Array: t.TypeAlias = npt.NDArray[np.uint8]
Uint16Array: t.TypeAlias = npt.NDArray[np.uint16]
Uint32Array: t.TypeAlias = npt.NDArray[np.uint32]
AnyArray: t.TypeAlias = npt.NDArray[t.Any]
IntpArray: t.TypeAlias = npt.NDArray[np.intp]
BoolArray: t.TypeAlias = npt.NDArray[np.bool_]
StrArray: t.TypeAlias = npt.NDArray[np.str_]


# ============================================================================ #
# Access package data


DATA_DIR = Path(__file__).parent / "data"


def get_data_path(filename: str, *, strict: bool = False) -> Path:
    """
    Construct a filepath to the data directory. If `strict` is True, a missing
    file with throw an error.
    """
    path = DATA_DIR / filename
    if strict and not path.exists():
        err = FileNotFoundError("Failed to find internal package file.")
        err.add_note("This is an internal error. Please report this.")
        raise err

    return path


# ============================================================================ #
# Create a lookup table


@njit(parallel=True)
def create_lookup_table(arr_lab: Float64Array, all_lab: Float64Array) -> Uint16Array:
    arr_lab_N = len(arr_lab)
    out = np.zeros(256**3, dtype=np.uint16)
    for i in prange(256**3):
        lab1 = all_lab[i]
        deltaes = np.zeros(arr_lab_N, dtype=np.float32)
        for j in prange(arr_lab_N):
            deltaes[j] = delta_e_cie2000(
                lab1[0],
                lab1[1],
                lab1[2],
                arr_lab[j][0],
                arr_lab[j][1],
                arr_lab[j][2],
            )
        out[i] = np.argmin(deltaes)

    return out


@njit(inline="always", cache=True)
def get_all_rgbs() -> Uint8Array:
    """
    Returns an array of all RGB triplets.
    """
    rgb = np.empty((256**3, 3), dtype=np.uint8)
    arr = np.arange(256, dtype=np.uint8).repeat(256**2)
    rgb[:, 0] = arr
    rgb[:, 1] = arr.reshape((-1, 256)).T.ravel()
    rgb[:, 2] = arr.reshape((-1, 256**2)).T.ravel()
    return rgb


@njit(cache=True)
def get_all_lab() -> Float64Array:
    """
    Returns a numpy array of the RGB colors in the LAB colorspace.
    """
    return rgb_to_lab(get_all_rgbs())
